from typing import Iterator
import multiprocess as mp
import numpy as np
import pandas as pd
import cooler
from cooltools.api.coverage import coverage

cooler.set_verbosity_level(2)


def generate_noise_pixels(
    bin_ids: np.array,
    n_contacts: int,
    chunksize: int
) -> Iterator[pd.DataFrame]:

    def _generate_chunk(n: int) -> pd.DataFrame:
        # Sample random pseudocontacts
        contacts = pd.DataFrame({
            "bin1_id": np.random.choice(bin_ids, n),
            "bin2_id": np.random.choice(bin_ids, n),
        })

        # Flip them to make them upper triangular
        mask = contacts["bin1_id"] > contacts["bin2_id"]
        tmp = contacts.loc[mask, "bin1_id"].copy()
        contacts.loc[mask, "bin1_id"] = contacts.loc[mask, "bin2_id"].copy()
        contacts.loc[mask, "bin2_id"] = tmp

        # Aggregate into pixels
        pixels = (
            contacts
            .groupby(["bin1_id", "bin2_id"])
            .size()
            .rename("count")
            .reset_index()
        )
        return pixels

    # Yield in chunks
    while n_contacts > 0:
        n = min(n_contacts, chunksize)
        yield _generate_chunk(n)
        n_contacts -= n


# Get the total number of contacts in the original dataset
binsize = 50_000
clr = cooler.Cooler(f"/net/projects/hic-atlas/spracklin-et-al-datasets/Unsynchronized.hg38.mapq_30.1000.mcool::resolutions/{binsize}")
n_contacts_total = clr.info["sum"]

# We don't want to inject noise into unmappable regions of the genome
with mp.Pool(32) as pool:
    cis_cov, tot_cov = coverage(clr, ignore_diags=2, map=pool.map)
trs_cov = tot_cov - cis_cov
good_bin_ids = np.where(tot_cov > 0)[0]


for downsample_perc in [25, 10]: # [95, 90, 85, 75, 50, 35, 10]:
    print(f"Refilling from {downsample_perc}\%")
    clr = cooler.Cooler(f"/net/projects/hic-atlas/test-data/Unsynchronized.hg38.mapq_30.downsample_{downsample_perc}.{binsize}.cool")
    bins = clr.bins()[:]

    # Make the noise map
    n_contacts_to_fill = n_contacts_total - clr.info["sum"]
    cooler.create_cooler(
        "/tmp/noise.cool",
        bins,
        generate_noise_pixels(good_bin_ids, n_contacts_to_fill, chunksize=100_000_000)
    )

    # Merge with the downsampled map
    cooler.merge_coolers(
        f"/net/projects/hic-atlas/test-data/Unsynchronized.hg38.mapq_30.downsample_{downsample_perc}+noise.{binsize}.cool",
        [f"/net/projects/hic-atlas/test-data/Unsynchronized.hg38.mapq_30.downsample_{downsample_perc}.{binsize}.cool", "/tmp/noise.cool"],
        mergebuf=100_000_000
    )

    assert cooler.Cooler(
        f"/net/projects/hic-atlas/test-data/Unsynchronized.hg38.mapq_30.downsample_{downsample_perc}+noise.{binsize}.cool"
    ).info["sum"] == n_contacts_total

